from .converter_helpers import (
    to_graphene,
    ConverterToGrapheneBase,
    ToGrapheneOptions,
)

from .errors import (
    Pydantic2GrapheneException,
    FieldNotSupported,
    InvalidType,
    InvalidListType,
    InvalidConfigClass,
)

from .version import VERSION


__version__ = VERSION


__all__ = (
    "__version__",
    "to_graphene",
    "ConverterToGrapheneBase",
    "ToGrapheneOptions",
    "Pydantic2GrapheneException",
    "FieldNotSupported",
    "InvalidType",
    "InvalidListType",
    "InvalidConfigClass",
)



import typing
import logging
import graphene
import pydantic
import inspect

from . import errors
from . import fields
from . import types

_module_cache = {}


def _get_pydantic_class_name(pydantic_model: pydantic.BaseModel) -> str:
    """
    >>> str(Human)
    "<class 'api.Human'>"
    """
    return str(pydantic_model).split("'")[1].split(".")[-1]


def _get_pydantic_fields(
    pydantic_model: pydantic.BaseModel,
) -> typing.List[pydantic.fields.ModelField]:
    """
    >>> str(Human)
    "<class 'api.Human'>"
    """
    return pydantic_model.__fields__.values()


class ToGrapheneOptions(pydantic.BaseModel):
    id_field_name: str = None

    extra_fields: typing.Mapping[str, types.GrapheneField] = {}

    exclude_fields: typing.Set[str] = set()

    class_name: str = None

    use_cache: bool = True

    @pydantic.validator("extra_fields")
    def validate_extra_fields(cls, value):
        if not value:
            return value

        for k, v in value.items():
            if not fields.is_graphene_type(v):
                raise errors.InvalidType(
                    f'Invalid field "{k}", is not a graphene filed'
                )

        return value


class ToGraphene:
    _cache: dict
    options: ToGrapheneOptions
    pydantic_model: pydantic.BaseModel
    graphene_type: types.GrapheneObjectType

    def __init__(
        self: object,
        pydantic_model: pydantic.BaseModel,
        graphene_type: types.GrapheneObjectType = graphene.ObjectType,
        options: typing.Union[ToGrapheneOptions, dict] = None,
        cache: dict = None,
    ):
        options = options or {}
        if not isinstance(options, ToGrapheneOptions):
            options = ToGrapheneOptions(**options)

        self.options = options
        self.pydantic_model = pydantic_model
        self.graphene_type = graphene_type
        self._cache = cache or _module_cache

    def _get_from_cache(
        self,
        pydantic_model: pydantic.BaseModel,
        graphene_type: types.GrapheneObjectType = graphene.ObjectType,
    ) -> types.GrapheneObjectType:
        cache_key = (pydantic_model, graphene_type)
        if not inspect.isclass(pydantic_model):  # when instance get the class
            cache_key = (pydantic_model.__class__, graphene_type)

        if self.options.use_cache and cache_key in self._cache:
            return (cache_key, self._cache[cache_key])

        return (cache_key, None)

    def _generate_class_name(self):
        _name = _get_pydantic_class_name(self.pydantic_model)

        if issubclass(self.graphene_type, graphene.InputObjectType):
            return f"{_name}InputGql"

        if issubclass(self.graphene_type, graphene.Interface):
            return f"{_name}InterfaceGql"

        return f"{_name}Gql"

    def _convert_to_graphene_field(
        self, pydantic_field: pydantic.fields.ModelField
    ):
        shape = pydantic_field.shape
        if fields.is_not_supported_shape(shape):
            raise errors.FieldNotSupported(pydantic_field.name)

        type_ = pydantic_field.type_
        type_args = getattr(type_, "__args__", [])

        if fields.is_list_shape(shape) and len(type_args):
            type_ = type_args[0]
            if len(type_args) > 1:
                logging.warn(
                    "%s, has multiple only the first type was used",
                    pydantic_field.name,
                )

        if fields.is_enum_type(type_):
            cache_key, cached_obj = self._get_from_cache(type_, graphene.Enum)
            if cached_obj:
                return cached_obj

            graphene_enum = graphene.Enum.from_enum(type_)
            self._cache[cache_key] = graphene_enum
            return graphene_enum

        if fields.is_pydantic_base_model(type_):
            if self.graphene_type == graphene.InputObjectType:
                obj_type = self.graphene_type
            else:
                obj_type = graphene.ObjectType

            return ToGraphene(type_, obj_type).convert()

        field = fields.get_grapehene_field_by_type(type_)
        if field:
            return field

        if fields.is_field_not_allowed_type(type_):
            raise errors.InvalidListType(
                "Lists must be type, e.g typing.List[int]"
            )

        outer_type_ = getattr(pydantic_field, "outer_type_", None)
        if outer_type_:
            field_outer = fields.get_grapehene_field_by_type(outer_type_)
            if field_outer:
                return field_outer

    def _get_graphene_field(self, pydantic_field: pydantic.fields.ModelField):
        args = {
            "required": pydantic_field.required,
            "default_value": pydantic_field.default,
        }
        field = self._convert_to_graphene_field(pydantic_field)
        if not field:
            raise errors.FieldNotSupported(pydantic_field.name)

        if fields.is_list_shape(pydantic_field.shape):
            if pydantic_field.required:
                return graphene.List(graphene.NonNull(field), **args)
            return graphene.List(field, **args)

        return graphene.Field(field, **args)

    def convert(self) -> types.GrapheneObjectType:

        cache_key, cached_obj = self._get_from_cache(
            self.pydantic_model, self.graphene_type
        )
        if cached_obj:
            return cached_obj

        graphene_attrs = {}
        for field in _get_pydantic_fields(self.pydantic_model):
            if field.name in self.options.extra_fields:
                continue

            if field.name in self.options.exclude_fields:
                continue

            graphene_attrs[field.name] = self._get_graphene_field(field)

        graphene_attrs.update(self.options.extra_fields)

        if self.options.id_field_name:
            graphene_attrs[self.options.id_field_name] = graphene.ID(
                required=True
            )

        class_name = self.options.class_name or self._generate_class_name()

        GrapheneClass = type(class_name, (self.graphene_type,), graphene_attrs)

        self._cache[cache_key] = GrapheneClass
        return GrapheneClass



import datetime
import typing
import ipaddress
import enum
import decimal
import uuid
import inspect

import graphene
import graphene.types.datetime
import pydantic.fields


_NOT_SUPPORTED_SHAPES = None


def _get_not_supported_shapes():
    global _NOT_SUPPORTED_SHAPES
    if _NOT_SUPPORTED_SHAPES:
        return _NOT_SUPPORTED_SHAPES

    _NOT_SUPPORTED_SHAPES = {
        pydantic.fields.SHAPE_MAPPING,
    }

    # pydantic pydantic<=1.8 does not have SHAPE_DICT and SHAPE_DEFAULTDICT
    try:
        _NOT_SUPPORTED_SHAPES.add(pydantic.fields.SHAPE_DICT)
        _NOT_SUPPORTED_SHAPES.add(pydantic.fields.SHAPE_DEFAULTDICT)
    except AttributeError:
        pass

    return _NOT_SUPPORTED_SHAPES


_LIST_SHAPES = None


def _get_list_shapes():
    global _LIST_SHAPES
    if _LIST_SHAPES:
        return _LIST_SHAPES

    _LIST_SHAPES = {
        pydantic.fields.SHAPE_LIST,
        pydantic.fields.SHAPE_TUPLE,
        pydantic.fields.SHAPE_TUPLE_ELLIPSIS,
        pydantic.fields.SHAPE_SEQUENCE,
        pydantic.fields.SHAPE_SET,
        pydantic.fields.SHAPE_FROZENSET,
    }

    # pydantic pydantic<=1.3 does not have SHAPE_ITERABLE
    try:
        _LIST_SHAPES.add(pydantic.fields.SHAPE_ITERABLE)
    except AttributeError:
        pass

    return _LIST_SHAPES


_TYPE_MAPPING = None


def _get_type_mapping():
    global _TYPE_MAPPING
    if _TYPE_MAPPING:
        return _TYPE_MAPPING

    _TYPE_MAPPING = {
        str: graphene.String,
        bool: graphene.Boolean,
        float: graphene.Float,
        int: graphene.Int,
        bytes: graphene.String,
        datetime.datetime: graphene.types.datetime.DateTime,
        typing.Pattern: graphene.String,
        ipaddress.IPv4Address: graphene.String,
        ipaddress.IPv4Interface: graphene.String,
        ipaddress.IPv4Network: graphene.String,
        ipaddress.IPv6Address: graphene.String,
        ipaddress.IPv6Interface: graphene.String,
        ipaddress.IPv6Network: graphene.String,
        decimal.Decimal: graphene.Float,
        uuid.UUID: graphene.String,
        pydantic.EmailStr: graphene.String,
        pydantic.NameEmail: graphene.String,
        pydantic.color.Color: graphene.String,
        pydantic.PaymentCardNumber: graphene.String,
        pydantic.AnyUrl: graphene.String,
        pydantic.AnyHttpUrl: graphene.String,
        pydantic.HttpUrl: graphene.String,
        pydantic.PostgresDsn: graphene.String,
        pydantic.RedisDsn: graphene.String,
        pydantic.UUID1: graphene.String,
        pydantic.UUID3: graphene.String,
        pydantic.UUID4: graphene.String,
        pydantic.UUID5: graphene.String,
        pydantic.SecretBytes: graphene.String,
        pydantic.SecretStr: graphene.String,
        pydantic.IPvAnyAddress: graphene.String,
        pydantic.IPvAnyInterface: graphene.String,
        pydantic.IPvAnyNetwork: graphene.String,
        pydantic.NegativeFloat: graphene.Float,
        pydantic.NegativeInt: graphene.Int,
        pydantic.PositiveFloat: graphene.Float,
        pydantic.PositiveInt: graphene.Int,
        pydantic.conbytes: graphene.String,
        pydantic.types.ConstrainedBytes: graphene.String,
        pydantic.condecimal: graphene.Float,
        pydantic.types.ConstrainedDecimal: graphene.Float,
        pydantic.confloat: graphene.Float,
        pydantic.types.ConstrainedFloat: graphene.Float,
        pydantic.conint: graphene.Int,
        pydantic.types.ConstrainedInt: graphene.Int,
        pydantic.constr: graphene.String,
        pydantic.types.ConstrainedStr: graphene.String,
    }

    # graphene does not support Date on versions:
    # ('1.X', '2.0')
    try:
        _TYPE_MAPPING[datetime.date] = graphene.types.datetime.Date
    except AttributeError:
        pass

    # graphene does not support Time on versions:
    # ('1.1.2', '1.1.1', '1.1', '1.0.2', '1.0.1', '1.0')
    try:
        _TYPE_MAPPING[datetime.time] = graphene.types.datetime.Time
    except AttributeError:
        pass

    # pydantic needs 'email-validator' to these fields
    try:
        _TYPE_MAPPING[pydantic.EmailStr] = graphene.String
        _TYPE_MAPPING[pydantic.NameEmail] = graphene.String
    except ImportError:
        pass

    # graphene does not support JSONString on versions:
    # ( 1.0>=, <=1.4.2 )
    try:
        _TYPE_MAPPING[pydantic.Json] = graphene.types.json.JSONString
    except AttributeError:
        pass

    return _TYPE_MAPPING


_LIST_FIELDS_NOT_TYPED = {
    list,
    tuple,
    dict,
    set,
    frozenset,
}


_ENUM_TYPE = (
    enum.Enum,
    enum.IntEnum,
)


def get_grapehene_field_by_type(type_):
    mapping = _get_type_mapping()

    if type_ in mapping:
        return mapping[type_]

    if inspect.isclass(type_):
        for pydantic_type in type_.mro():
            if pydantic_type in mapping:
                return mapping[pydantic_type]


def is_field_not_allowed_type(type_) -> bool:
    return type_ in _LIST_FIELDS_NOT_TYPED


def is_enum_type(type_) -> bool:
    return inspect.isclass(type_) and issubclass(type_, _ENUM_TYPE)


def is_list_shape(shape) -> bool:
    return shape in _get_list_shapes()


def is_not_supported_shape(shape) -> bool:
    return shape in _get_not_supported_shapes()


def is_pydantic_base_model(type_) -> bool:
    return inspect.isclass(type_) and issubclass(type_, pydantic.BaseModel)


def is_graphene_type(type_) -> bool:
    return "graphene.types" in repr(type_)



import typing
import graphene
import pydantic
import inspect

from . import errors
from . import fields
from . import converter
from . import types

ToGrapheneOptions = converter.ToGrapheneOptions


def to_graphene(
    pydantic_model: pydantic.BaseModel,
    graphene_type: types.GrapheneObjectType = graphene.ObjectType,
    options: typing.Union[ToGrapheneOptions, dict] = None,
) -> types.GrapheneObjectType:
    return converter.ToGraphene(
        pydantic_model, graphene_type, options
    ).convert()


class ConverterToGrapheneBase:
    @classmethod
    def as_class(
        cls, graphene_type: types.GrapheneObjectType = None
    ) -> types.GrapheneObjectType:
        Config = getattr(cls, "Config", None)
        if not inspect.isclass(Config):
            raise errors.InvalidConfigClass("Config is invalid")

        model = getattr(Config, "model", None)
        if not model:
            raise errors.InvalidConfigClass(
                'Config is missing "model" property'
            )

        options = {k: v for k, v in vars(Config).items()}
        options["extra_fields"] = {
            attr: value
            for attr, value in vars(cls).items()
            if fields.is_graphene_type(value)
        }

        params = {
            "pydantic_model": model,
            "options": options,
        }
        graphene_type = graphene_type or getattr(Config, "graphene_type", None)
        if graphene_type:
            params["graphene_type"] = graphene_type

        return to_graphene(**params)



class Pydantic2GrapheneException(Exception):

    pass


class FieldNotSupported(Pydantic2GrapheneException):
    pass


class InvalidType(Pydantic2GrapheneException):
    pass


class InvalidListType(FieldNotSupported):
    pass


class InvalidConfigClass(Pydantic2GrapheneException):
    pass



import pathlib
import setuptools
import re


description = (
    "Easy way to convert pydantic2graphene models to graphene objects."
)
long_description = description

ROOT_DIR = pathlib.Path(__file__).resolve().parent
try:
    long_description = (ROOT_DIR / "README.md").read_text()
except FileNotFoundError:
    pass

version_txt = (ROOT_DIR / "pydantic2graphene" / "version.py").read_text()
version = re.search(r"v\d+(\.\d+)+", version_txt).group()

requirements = [
    "graphene>=1.1,<=2.1.8",
    "pydantic>=1.0,<=1.8.1",
]

github_url = "https://github.com/lfvilella/pydantic2graphene"

setuptools.setup(
    name="pydantic2graphene",
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX :: Linux",
        "Environment :: Console",
        "Environment :: MacOS X",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet",
    ],
    keywords=["python", "api", "graphql", "graphene", "pydantinc"],
    author="Luis Felipe Vilella",
    author_email="vilella.luisfelipe@gmail.com",
    url=github_url,
    download_url=f"{github_url}/archive/{version}.tar.gz",
    license="MIT",
    packages=["pydantic2graphene"],
    package_data={"pydantic2graphene": ["py.typed"]},
    python_requires=">=3.6,<=3.9",
    install_requires=requirements,
)
